#include "tools.h"
#include <filesystem> //cpp17

/*
    http://docs.opencv.org/2.4/doc/tutorials/imgproc/gausian_median_blur_bilateral_filter/gausian_median_blur_bilateral_filter.html
    https://learnopencv.com/edge-detection-using-opencv/
    https://learnopencv.com/contour-detection-using-opencv-python-c/
*/

void Tools::print_ruler()
{
	std::cout << "######################################" << std::endl;
}

void Tools::maybeAppendSepToDir(cv::String& d)
{
	const char SEP = std::filesystem::path::preferred_separator;
	if( d.back() != SEP )
		d += SEP;
}

bool Tools::check(const std::vector<cv::String>& x)
{
	bool checkAll = true;
	cv::String name, path;

	for (uint i{0}; i<x.size()-1; i++)
	{
		name = x[i];
		path = x[++i];

		bool check = std::filesystem::exists(path);
		if (!check)
			std::cerr << name << ": " << path << " [not found!] " << std::endl;
		else
			std::cout << name << ": " << path << std::endl;

		checkAll &= check;
	}

	return checkAll;
}

std::string Tools::mNzero(int i)
{
	std::ostringstream convert;
	convert << i ;
	std::string numberString = convert.str();
	std::string newNumberString = std::string(10 - numberString.length(), '0') + numberString;
	return newNumberString;
}
std::string Tools::getFileName(int i)
{
	std::string fileName = std::string ( std::string ("/data/") + mNzero(i) + std::string (".jpg"));
	return fileName;
}

std::vector<std::vector<cv::Point>> Tools::applyCannyEdgeAndCalcCountours(cv::Mat imgSource, double threshholdMin, double threshholdMax, int apertureSize)
{
    //TODO: check if clone is necessary or if it wastes ressources by doing unneccessary copies of frames

    //! make a copy
	cv::Mat imgBinary = imgSource.clone();

    //! apply binary filter not reduce noise
    // > method to threshold important changes
    int binaryThreshold = 80; //TODO as parameter
    imgBinary = imgSource > binaryThreshold;

    //! smooth with gaussian filter to suppress no connected lines
    //add a gaussian filter
    cv::Mat imgSmoothed = imgBinary.clone();
    int exclusPara3 = 3; //TODO as parameter
    cv::GaussianBlur(imgBinary,imgSmoothed,cv::Size(exclusPara3,exclusPara3),0);  //0 = BORDER_CONSTANT

    //! detect edges using canny
    cv::Mat imgCannyEdge;
    cv::Canny( imgSmoothed, imgCannyEdge, threshholdMin, threshholdMax, apertureSize );

    //! find contours
    //  RETR_EXTERNAL only extreme outer contours > hierarchy[i][2]=hierarchy[i][3]=-1
    //  RETR_LIST no hierarchy
    //  RETR_TREE full hierarchy of nested contours
    std::vector<std::vector<cv::Point> > contours;
    std::vector<cv::Vec4i> hierarchy;
    cv::findContours( imgCannyEdge, contours, hierarchy, cv::RETR_EXTERNAL, cv::CHAIN_APPROX_SIMPLE, cv::Point(0, 0) );

#define MC_SHOW_STEP_ANALYSE
#ifdef  MC_SHOW_STEP_ANALYSE
    // Draw contours on extra mat
    cv::Mat imgContour = cv::Mat::zeros( imgCannyEdge.size(), CV_8UC3 );
    imgContour =  cv::Scalar(255,255,255); //fill the picture
    cv::RNG rng(232323);
    for( size_t i = 0; i< contours.size(); i++ )
    {
        //random color
    	cv::Scalar color = cv::Scalar( rng.uniform(0, 255), rng.uniform(0,255), rng.uniform(0,255) );
        //contour
    	cv::drawContours( imgContour, contours, i, color, 1, cv::LINE_AA, hierarchy, 0, cv::Point() );
    }

    // bgslibrary-bgslib_qtgui_2.0.0
    cv::String outpath = "./";
    std::ostringstream convert;
    convert << outpath << "test_countour_canny_edge.jpg";
    cv::imwrite(convert.str().c_str(), imgContour);

#endif

    return contours;
  }




  cv::Mat Tools::cropImageCircle(cv::Mat image, int x, int y, int r)
  {

    // Hough circle
  	cv::Vec3f circ(x,y,r);

    // Draw the mask: white circle on black background
  	cv::Mat1b mask(image.size(), uchar(0));
  	cv::circle(mask, cv::Point(circ[0], circ[1]), circ[2], cv::Scalar(255), cv::FILLED);

    //NO CROP NEEDED

    // Compute the bounding box
    //Rect bbox(circ[0] - circ[2], circ[1] - circ[2], 2 * circ[2], 2 * circ[2]);

    // Create a black image
  	cv::Mat res = cv::Mat::zeros( image.size(), CV_8UC3 );
    res = cv::Scalar(0,0,0); //fill with black color //needed? isnt it initd with zeros above?


    // Copy only the image under the white circle to black image
    image.copyTo(res, mask);

    // Crop according to the roi
    //res = res(bbox);

    return res;
  }

  int Tools::checkFileCorrupted(std::string filename)
  {
    int retVal = 1;//is as long Corrupted as we dont find it in another way
    char command[] =  "identify ";
    const char *fileNameArr = filename.c_str();
    char command2[] = " > /dev/null 2>&1 ; echo $?";
    char *cmd = new char[std::strlen(command)+std::strlen(fileNameArr) + std::strlen(command2) +1];
    //we put all peaces together
    std::strcpy(cmd,command);
    std::strcat(cmd,fileNameArr);
    std::strcat(cmd,command2);

    //std:: cout <<"cmd is: "<< cmd<< std::endl;
    //execute the command
    std::string result = Tools::execCMD((const char* ) cmd);

    //std:: cout <<"results is: "<< result<< std::endl;

    if(!result.empty())
    {
    	try {
    		retVal = std::stoi(result);
    	}
    	catch(std::invalid_argument& e)
    	{
    		std::cerr <<"wrong argument for stoi"<<std::endl;
            // if no conversion could be performed
    	}
    	catch (const std::exception& e)
    	{
    		std::cerr <<"error during use stoi"<<std::endl;
    	}

    }//end if
    return retVal;
  }


  std::string Tools::execCMD(const char* cmd) {
  	std::array<char, 128> buffer;
  	std::string result;
  	std::shared_ptr<FILE> pipe(popen(cmd, "r"), pclose);
  	if (!pipe) throw std::runtime_error("popen() failed!");
  	while (!feof(pipe.get())) {
  		if (fgets(buffer.data(), 128, pipe.get()) != nullptr)
  			result += buffer.data();
        //std::cout<<result<<std::endl;
  	}
    //std::cout<<"call:       " <<result<<std::endl;
  	return result;
  }



// bool checkDirs(std::map<std::string, std::string> dirs)
// {
// 	bool ret = false;
// 	for(const auto& [k,v] : dirs) {
// 		if (!std::filesystem::exists(v))
// 		{
// 			// bool check = mkdir(dirname,0777);
// 			if(std::filesystem::create_directories(v))
// 			{
// 				std::cout << k << ": " << v << " [created!] " << std::endl;
// 			}
// 			else
// 			{
// 				std::cerr << k << ": " << v << " [not found + cant create it!] " << std::endl;
// 				ret = true;
// 			}
// 		}
// 		else
// 		{
// 			cout << k << ": " << v << std::endl;
// 		}
// 	}

// 	// add '/' if its missing
// 	//path p = outputDir / "_convex_hull.jpg";
// 	print_ruler();
// 	return ret;
// }
